; ======================================
; Depack routine.
;
; Includes packed data at the end of the file.
; Set LOADADDRESS and STARTADDRESS as required.
; ======================================

DEPACKER        = $0100                 ; Fits in the stack.
LOADADDRESS     = $0810
STARTADDRESS    = $0810

dest    = $60
src     = $62
string  = $64

        processor 6502

        org $0801
; Basic tokens for the line '2001 SYS2064ZED'
        .byte $0B,$08,$D1,$07,$9E,$32,$30,$36,$34,$5a,$45,$44,$00,$00

        org $0810

; Clear screen from right to left. Clever, eh?
        ldy #39
spalsh_nextframe:
        lda #$fa
splash_waitraster:
        cmp $d012
        bne splash_waitraster
        lda #$00
        sta src
        sta dest
        lda #$04
        sta src+1
        lda #$d8
        sta dest+1
        ldx #25
spalsh_nextline:
        lda #$a0
        sta (src),y
        lda $d020
        sta (dest),y
        lda src
        clc
        adc #40
        sta src
        sta dest
        lda src+1
        adc #0
        sta src+1
        adc #$d4
        sta dest+1
        dex
        bne spalsh_nextline
        dey
        bpl spalsh_nextframe

        sei
        lda #$34
        sta $01
        ldy #DEPACKLEN
copydepacker:
        lda depackstart-1,y
        sta DEPACKER-1,y
        dey
        bne copydepacker

; Copy to top of mem. src and dest are reversed in their meaning!
        sty src
        sty src+1
        ldx #DATAPAGES
copyup_loop:
        dec copyup_innerloop+2
        dec src+1
copyup_innerloop:
        lda datastart+256*DATAPAGES,y
        sta (src),y
        iny
        bne copyup_innerloop
        dex
        bne copyup_loop
        lda #<LOADADDRESS
        sta dest
        lda #>LOADADDRESS
        sta dest+1
        jmp DEPACKER

; Depacker.
depackstart:
        rorg DEPACKER

depackloop:
        jsr getbyte
        lsr
        beq depackready
        bcs compressedblock

; Uncompressed block.
        tax
uncompressedloop:
        jsr getbyte
        jsr putbyte
        dex
        bne uncompressedloop
        beq depackloop

; RLE or string block.
compressedblock:
        lsr
        tax
        bcs stringblock

; RLE block.
        jsr getbyte
rleloop:
        jsr putbyte
        dex
        bne rleloop
        beq depackloop

stringblock:
        jsr getbyte
        clc
        adc dest
        sta string
        php
        jsr getbyte
        plp
        adc dest+1
        sta string+1
stringblockloop:
        lda (string),y
        inc string
        bne getstringbyteok
        inc string+1
getstringbyteok:
        jsr putbyte
        dex
        bne stringblockloop
        beq depackloop

depackready:
        ldx #$ff
        txs
        lda #$37
        sta $01
        cli
        jmp STARTADDRESS

getbyte:
        lda (src),y
        inc src
        bne getbyteok
        inc src+1
getbyteok:
        rts

putbyte:
        sta (dest),y
        inc dest
        bne putbyteok
        inc dest+1
putbyteok:
        rts

        rend

DEPACKLEN       = *-depackstart


datastart:
        incbin packed.bin
dataend:
DATAPAGES       = >(dataend-datastart+$ff)
