; ======================================
; Part of Odin Tracker 1.12 for C64 by Zed on 20 Mar 2001.
;
; Editor's player.
; ======================================

        jmp player_init
        jmp player_play

; ======================================
; Damn keyjazz hack.
; ======================================

player_keyjazznote:             .byte 0
player_keyjazzinst:             .byte 0
player_previouskeyjazznote:     .byte 0

; ======================================
player_istrument_play:
; Run keyjazz.
; ======================================
;  Input: A: note ($ff means no key pressed)
;         Y: instrument
; ======================================
; Test always in channel 0.
        ldx #0
        sta player_keyjazznote
        sty player_keyjazzinst
        cmp #$ff
        bne pip_normalnote
        lda #$fe
        sta chn_gateon,x
pip_nonewnote1:
        jmp pip_nonewnote
pip_normalnote:
; If key was released, always start the new note.
        pha
        lda chn_gateon,x
        lsr
        pla
        bcc pip_newnote
; If the same key is held, do not start the sound again.
        cmp player_previouskeyjazznote
        beq pip_nonewnote1
pip_newnote:
; Clear player data.
        lda #$00
        tay
pip_clear:
        sta playercleardata,y
        iny
        cpy #PLAYERCLEARDATASIZE2
        bne pip_clear
        lda #$0f
        sta globalvolume
        lda player_keyjazznote
        sta chn_note,x
        lda player_keyjazzinst
        sta chn_inst,x
        tay
; Copy changeable instrument parameters to channel.
        lda INSTRUMENTS_AD,y            ; Copy AD.
        sta chn_ad,x
        lda INSTRUMENTS_SR,y            ; Copy SR.
        sta chn_sr,x
        lda INSTRUMENTS_WAVETABLESTART,y ; Wave table start index.
        sta chn_waveidx,x
        lda INSTRUMENTS_ARPTABLESTART,y ; Arpeggio table start index.
        sta chn_arpidx,x
        lda INSTRUMENTS_VIBDELAY,y      ; Copy vibrato delay.
        sta chn_vibdelay,x
        lda INSTRUMENTS_PULSEWIDTH,y    ; Copy pulse width.
        jsr set_pulsewidth
        lda INSTRUMENTS_PULSESPEED,y    ; Copy pulse speed.
        sta chn_plsspeed,x
        lda INSTRUMENTS_PULSELIMITS,y   ; Copy pulse limits.
        jsr set_pulselimits
; TODO init filter to operate on channel 0 !!!
; The user should be able to set up filter if he needs.
; The code below is not OK, because it always enable filters,
; which is not what we need.
; So for now, no filtering in keyjazz mode. Sorry.
;        lda INSTRUMENTS_FILTERTABLESTART,y  ; Filter table start index.
;        sta filter_idx
;        lda INSTRUMENTS_FILTERTABLEEND,y ; Copy filter table end.
;        sta filter_end
;        lda INSTRUMENTS_FILTERTABLELOOP,y ; Copy filter table loop.
;        sta filter_loop
;        lda #$f1
;        sta filter_input
;        lda #$10
;        sta filter_mode
; Reset vibrato and pulse width modulation.
        lda #$00
        sta chn_plsdir,x
        sta chn_vibpos,x
; Set note.
        lda chn_note,x
        clc
        adc chn_transpose,x
        jsr note2notefreq
; Set gate on.
        lda #$ff
        sta chn_gateon,x

pip_nonewnote:
        lda #$00
        sta chn_vibdepth,x
        jsr process_instrument
        lda chn_vibdepth,x
        beq pip_vibinactive
        jsr calcvibrato
        lda chn_vibpos,x
        clc
        adc chn_vibspeed,x
        sta chn_vibpos,x
        jmp pip_vibdone
pip_vibinactive:
        lda chn_freqlo,x
        sta chn_finfreqlo,x
        lda chn_freqhi,x
        sta chn_finfreqhi,x
pip_vibdone:

; Process filter table.
        ldy filter_idx
        lda FILTERTABLE,y
        sta filter_cutoff
; Advance filter table index.
        iny
        cpy filter_end
        bcc pip_filteridxok
        beq pip_filteridxok
        ldy filter_loop
pip_filteridxok:
        sty filter_idx

; Dump everything into the SID chip.
        ldx #$00
        ldy sidregindex,x
        lda chn_plswidthlo,x
        sta $d402,y
        lda chn_plswidthhi,x
        sta $d403,y
        lda chn_finfreqlo,x
        sta $d400,y
        lda chn_finfreqhi,x
        sta $d401,y
        lda chn_waveform,x
        and chn_gateon,x
        sta $d404,y
        lda chn_ad,x
        sta $d405,y
        lda chn_sr,x
        sta $d406,y
        lda filter_cutoff
        sta $d416
        lda filter_input
        sta $d417
        lda globalvolume
        ora filter_mode
        sta $d418
        lda chn_note,x
        sta player_previouskeyjazznote
        rts

; ======================================
player_init:
; This is used outside the editor.
; ======================================
;  Input: A: song number
; ======================================
        tax
        lda #$00
        sta chn_muted+0
        sta chn_muted+1
        sta chn_muted+2
        sta player_singlepattern
        lda SONGSTARTTABLE,x

; ======================================
player_initforeditor:
; ======================================
;  Input: A: start order number
;         X: start pattern row
;         chn_muted[]
;         player_singlepattern
; ======================================
        pha
        lda #$00
        tay
pi_clear:
        sta playercleardata,y
        iny
        cpy #PLAYERCLEARDATASIZE
        bne pi_clear
        lda #$80                        ; Force a pattern jump.
        sta forcenewpattern
; Restore order position.
        pla
        sta ordernumber
        sta nextordernumber
        stx firsttrackrow               ; Able to start playing from any row.
; Init speed and speedcounter.
        ldy #6
        sty speed
        dey
        sty speedcounter
; Init hard restart counters.
        lda #2
        sta chn_hardrestart+0
        sta chn_hardrestart+1
        sta chn_hardrestart+2
; Set global volume to maximum.
        lda #$0f
        sta globalvolume

; ======================================
player_stop:
; ======================================
; Shut up SID.
        lda #$00
        ldy #$18
ps_resetsidloop:
        sta $d400,y
        dey
        bpl ps_resetsidloop
        rts

; ======================================
player_play:
; Call this in every frame.
; ======================================
; Play 1 tick of music.
; ======================================
; Update the mod3 counter which is used for the arpeggio effect.
        dec mod3counter
        bpl pt_mod3counterdone
        lda #2
        sta mod3counter
pt_mod3counterdone:
; Speed 0 stops song, but continues effects.
        lda speed
        beq pt_processeffects           ; Do effects only.
; Update speed counter and see if it is time for new trackrow.
        inc speedcounter
        lda speedcounter
        cmp speed
        bne pt_processeffects           ; Do effects only.
; Start new row.
        lda #$00                        ; Reset speed counter.
        sta speedcounter
        lda forcenewpattern             ; Fetching a new pattern is forced at
        bmi pt_newpatternforced         ; start of tune and by order jump and pattern break.
        inc trackrow                    ; Advance one row.
        lda trackrow                    ; Check if we reached the end of pattern.
        cmp #64
        bmi pt_pattern_ok
pt_newpatternforced:
        jsr start_newpattern            ; !!!! Expand inline!

pt_pattern_ok:
; Create offset into track. trackrow3=3*trackrow
        lda trackrow
        asl
        adc trackrow
        sta trackrow3
; Fetch new lines from tracks.
        ldx #$00
pt_fetchnextchn:
        lda chn_trackptrlo,x
        sta player_trackptr
        lda chn_trackptrhi,x
        sta player_trackptr+1
        jsr fetchrow                    ; !!!! Expand inline!
        inx
        cpx #$03
        bne pt_fetchnextchn

; Process running effects for all channels.
pt_processeffects:
        ldx #$00
ptpe_nextchn:
; Suppose there is no vibrato.
; Effect and instrument vibrato can override this.
        lda #$00
        sta chn_vibdepth,x
; First check if there is an instrument. If there is none,
; there is no instrument wavetable/arpeggio/vibrato to do.
        lda chn_inst,x
        beq ptpe_noinst
; This is done before effects, so that they can override this.
        jsr process_instrument

ptpe_noinst:
; Now process track effects.
        lda chn_effect,x
        beq pt_processeffectsdone
        asl
        tay
        lda effectpointers,y
        sta pt_processeffects0+1
        lda effectpointers+1,y
        sta pt_processeffects0+2
pt_processeffects0:
        jsr $ffff                       ; !!!! Self-modifying.
pt_processeffectsdone:
; Add slide frequency, which is negative when sliding down.
        lda chn_slideactive,x
        beq ptpe_slideinactive
        lda chn_freqlo,x
        clc
        adc chn_slidefreqlo,x
        sta chn_freqlo,x
        lda chn_freqhi,x
        adc chn_slidefreqhi,x
        sta chn_freqhi,x
ptpe_slideinactive:
; Do vibrato if necessary.
        lda chn_vibdepth,x
        beq ptpe_vibinactive
        jsr calcvibrato
        lda chn_vibpos,x
        clc
        adc chn_vibspeed,x
        sta chn_vibpos,x
        jmp ptpe_vibdone
ptpe_vibinactive:
        lda chn_freqlo,x
        sta chn_finfreqlo,x
        lda chn_freqhi,x
        sta chn_finfreqhi,x
ptpe_vibdone:
        inx
        cpx #3
        bne ptpe_nextchn

; Process filter table.
        ldy filter_idx
        beq pt_filterdone
        lda FILTERTABLE,y
        sta filter_cutoff
; Advance filter table index.
        iny
        cpy filter_end
        bcc pt_filteridxok
        beq pt_filteridxok
        ldy filter_loop
pt_filteridxok:
        sty filter_idx
pt_filterdone:

; Do hardrestart stuff.
; !!!! TODO: shit happens if we have to look ahead into the next pattern.
; This should work but it is damn waste of time.
        lda forcenewpattern             ; Fetching a new pattern is forced at
        bmi pthr_newpatternforced       ; start of tune and by order jump and pattern break.
        ldx trackrow
        inx
        txa
        and #$3f
        stx nexttrackrow
        bne pthr_pattern_ok
pthr_newpatternforced:
        lda firsttrackrow               ; For pattern break, this is the next row.
        sta nexttrackrow
        jsr look_newpattern             ; !!!! Expand inline!

pthr_pattern_ok:
; Create offset into track. Y=3*nexttrackrow
        lda nexttrackrow
        asl
        adc nexttrackrow
        tay
        ldx #$00
pthr_nextchn:
; Is it time for hardrestart?
        lda speedcounter
        clc
        adc chn_hardrestart,x
        cmp speed
        bcc pthr_nohardrestart
        lda chn_trackptrlo,x
        sta player_trackptr
        lda chn_trackptrhi,x
        sta player_trackptr+1
        lda (player_trackptr),y
        and #$7f
        beq pthr_nohardrestart          ; No note -> no hardrestart.
        cmp #97
        beq pthr_nohardrestart          ; Note off -> no hardrestart.
; We will have a new note, but if we'll also have slide to note (effect 3)
; never do hardrestart.
        lda (player_trackptr),y
        bmi pthr_hardrestart            ; Can't be effect 3 -> do hardrestart.
        iny
        lda (player_trackptr),y
        dey
        and #$e0
        cmp #$60
        beq pthr_nohardrestart          ; This is effect 3 -> no hardrestart.
pthr_hardrestart:
; Clear AD/SR and waveform to zero.
; Maybe some other ways would result in better hardrestart.
; Tell me if you have any good ideas.
        lda #$00
        sta chn_waveform,x
        sta chn_ad,x
        sta chn_sr,x
pthr_nohardrestart:
        inx
        cpx #$03
        bne pthr_nextchn

; Dump everything into the SID chip.
        ldx #2
pt_dump2sid:
        ldy sidregindex,x
        lda chn_plswidthlo,x
        sta $d402,y
        lda chn_plswidthhi,x
        sta $d403,y
        lda chn_finfreqlo,x
        sta $d400,y
        lda chn_finfreqhi,x
        sta $d401,y
        lda chn_muted,x
        beq ptds_channeon
        lda #$00
        sta $d404,y
        sta $d405,y
        sta $d406,y
        beq ptds_channedone
ptds_channeon:
        lda chn_waveform,x
        and chn_gateon,x
        sta $d404,y
        lda chn_ad,x
        sta $d405,y
        lda chn_sr,x
        sta $d406,y
ptds_channedone:
        dex
        bpl pt_dump2sid
; Update filter parameters.
        lda filter_cutoff
        sta $d416
        lda filter_input
        sta $d417
        lda globalvolume
        ora filter_mode
        sta $d418
        rts

; ======================================
process_instrument:
; Process instrument arpeggio, vibrato, etc.
; ======================================
;  Input: X: channel number
; Output: chn_*, !!!! many
;   Uses: chn_*, !!!! many
; ======================================
; Process wave table.
        ldy chn_waveidx,x
        lda WAVETABLE,y
        sta chn_waveform,x
; Advance wavetable index.
        iny
        tya
        ldy chn_inst,x
        cmp INSTRUMENTS_WAVETABLEEND,y
        bcc pi_waveidxok
        beq pi_waveidxok
        lda INSTRUMENTS_WAVETABLELOOP,y
pi_waveidxok:
        sta chn_waveidx,x

; If slide to note is active, do not do arpeggio.
        lda chn_effect,x
        cmp #3
        beq pi_skiparpeggio

; Process arpeggio table.
        ldy chn_arpidx,x
        lda ARPEGGIOTABLE,y
; Arpeggio table entry is < $80 -> add this note to base (relative arpeggio),
; otherwise use entry-$80 (absolute arpeggio)
        bmi pi_arptableabs
        clc
        adc chn_note,x
        adc chn_transpose,x
pi_arptableabs:
        and #$7f
; Advance arpeggio table index.
        pha
        iny
        tya
        ldy chn_inst,x
        cmp INSTRUMENTS_ARPTABLEEND,y
        bcc pi_arpidxok
        beq pi_arpidxok
        lda INSTRUMENTS_ARPTABLELOOP,y
pi_arpidxok:
        sta chn_arpidx,x
        pla
        jsr note2freq                   ; Set frequency for this note.

; Process instrument vibrato.
pi_skiparpeggio:
        lda chn_vibdelay,x
        bne pi_vibdelayed
        ldy chn_inst,x
        lda INSTRUMENTS_VIBDEPTH_SPEED,y
        and #$f0
        sta chn_vibdepth,x
        beq pi_vibdone                  ; If depth==0, then no vibrato.
        lda INSTRUMENTS_VIBDEPTH_SPEED,y
        and #$0f
        sta chn_vibspeed,x
        jmp pi_vibdone
pi_vibdelayed:
        dec chn_vibdelay,x
pi_vibdone:

; Do pulse width modulation.
        lda chn_plsspeed,x
        beq pi_pulsedone
        lda chn_plsdir,x
        beq pi_pulseup
; Decrement pulse width. If lower limit is reached, reverse direction.
        lda chn_plswidthlo,x
        sec
        sbc chn_plsspeed,x
        sta chn_plswidthlo,x
        lda chn_plswidthhi,x
        sbc #0
        cmp chn_plslimitdown,x
        bpl pi_pulsedownok
        lda #$00
        sta chn_plsdir,x
        sta chn_plswidthlo,x
        lda chn_plslimitdown,x
pi_pulsedownok:
        sta chn_plswidthhi,x
        rts
; Increment pulse width. If upper limit is reached, reverse direction.
pi_pulseup:
        lda chn_plswidthlo,x
        clc
        adc chn_plsspeed,x
        sta chn_plswidthlo,x
        lda chn_plswidthhi,x
        adc #0
        cmp chn_plslimitup,x
        bmi pi_pulseupok
        beq pi_pulsedownok
        lda #$80
        sta chn_plsdir,x
        lda #$ff
        sta chn_plswidthlo,x
        lda chn_plslimitup,x
pi_pulseupok:
        sta chn_plswidthhi,x
pi_pulsedone:
        rts

; ======================================
effect00:
; ======================================
; Do nothing. This is in fact never called.
; ======================================
        rts

; ======================================
effect01:
; ======================================
; Slide down by param if param<$80, otherwise up by param-$80.
; ======================================
; Indicate that slide is active.
        lda #$80
        sta chn_slideactive,x
; Calculate slide adder.
        lda chn_effectpar,x
        and #$7f
; Multiply parameter by 16 to get slide adder.
        tya
        asl
        asl
        asl
        asl
        sta player_vibratotemp
        tya
        lsr
        lsr
        lsr
        lsr
        sta player_vibratotemp+1
        ldy chn_effectpar,x
        bmi effect01up
; Slide down - subtract from frequency.
        lda chn_slidefreqlo,x
        sec
        sbc player_vibratotemp
        sta chn_slidefreqlo,x
        lda chn_slidefreqhi,x
        sbc player_vibratotemp+1
        sta chn_slidefreqhi,x
        rts
; Slide up - add to frequency.
effect01up:
        lda chn_slidefreqlo,x
        clc
        adc player_vibratotemp
        sta chn_slidefreqlo,x
        lda chn_slidefreqhi,x
        adc player_vibratotemp+1
        sta chn_slidefreqhi,x
        rts

; ======================================
effect02:
; ======================================
; Set pulse width.
; Parameter is bits 11..4 of pulse width.
; ======================================
        lda chn_effectpar,x
        jmp set_pulsewidth

; ======================================
effect03:
; ======================================
; Slide to note. Damn hack.
; ======================================
        ldy chn_effectpar,x
; Multiply parameter by 16 to get slide adder.
        tya
        asl
        asl
        asl
        asl
        sta player_vibratotemp
        tya
        lsr
        lsr
        lsr
        lsr
        sta player_vibratotemp+1
        jsr freqcmp
        bcc effect03up
; Slide down.
        lda chn_freqlo,x
        sec
        sbc player_vibratotemp
        sta chn_freqlo,x
        lda chn_freqhi,x
        sbc player_vibratotemp+1
        sta chn_freqhi,x
; Check if we have shot past the destination frequency.
        jsr freqcmp
        bcc effect03stop                ; Reached, stop.
        rts                             ; Still not reached.
; Slide up.
effect03up:
        lda chn_freqlo,x
        clc
        adc player_vibratotemp
        sta chn_freqlo,x
        lda chn_freqhi,x
        adc player_vibratotemp+1
        sta chn_freqhi,x
; Check if we have shot past the destination frequency.
        jsr freqcmp
        bcs effect03stop                ; Reached, stop.
        rts                             ; Still not reached.
effect03stop:
        lda chn_notefreqlo,x
        sta chn_freqlo,x
        lda chn_notefreqhi,x
        sta chn_freqhi,x
        rts

; ======================================
effect04:
; ======================================
; Vibrato.
; High nybble of parameter is depth, low is speed.
; ======================================
        lda chn_effectpar,x
        and #$f0
        sta chn_vibdepth,x
        lda chn_effectpar,x
        and #$0f
        sta chn_vibspeed,x
        rts

; ======================================
effect05:
; ======================================
; Set pulse speed.
; ======================================
        lda chn_effectpar,x
        sta chn_plsspeed,x
        rts

; ======================================
effect06:
; ======================================
; Set pulse limits.
; ======================================
        lda chn_effectpar,x
        jsr set_pulselimits
        rts

; ======================================
effect07:
; ======================================
; Set Attack/Decay.
; ======================================
        lda chn_effectpar,x
        sta chn_ad,x
        rts

; ======================================
effect08:
; ======================================
; Set Sustain/Release.
; ======================================
        lda chn_effectpar,x
        sta chn_sr,x
        rts

; ======================================
effect09:
; ======================================
; Set waveform.
; ======================================
        lda chn_effectpar,x
        sta chn_waveform,x
        rts

; ======================================
effect0a:
; ======================================
; Arpeggio. (Protracker-like.)
; ======================================
        ldy mod3counter
        beq effect0aadd2                ; Counter at 0 -> use note 2.
        dey
        bne effect0aadd0                ; Counter at 2 -> use base note.
        lda chn_effectpar,x             ; Counter at 1 -> use note 1.
        lsr
        lsr
        lsr
        lsr
        clc
        adc chn_note,x
        bcc effect0adone
effect0aadd0:
        lda chn_note,x
        clc
        bcc effect0adone
effect0aadd2:
        lda chn_effectpar,x
        and #$0f
        clc
        adc chn_note,x
effect0adone:
        adc chn_transpose,x
        jmp note2freq

; ======================================
effect0b:
; ======================================
; Jump to order.
; ======================================
        lda chn_effectpar,x
        sta nextordernumber
        lda #$80
        sta forcenewpattern
        rts

; ======================================
effect0c:
; ======================================
; Set filter cutoff frequency.
; Parameter is 8 MSBs of frequency.
; ======================================
        lda chn_effectpar,x
        sta filter_cutoff
        rts

; ======================================
effect0d:
; ======================================
; Pattern break.
; ======================================
        lda chn_effectpar,x
        sta firsttrackrow
        lda #$80
        sta forcenewpattern
        rts

; ======================================
effect0e:
; ======================================
; Filter resonance/input control.
; High nybble is resonance, bit 0,1,2 enable filter for voice 1,2,3.
; ======================================
        lda chn_effectpar,x
        sta filter_input
        rts

; ======================================
effect0f:
; ======================================
; Parameter: $00-$7f: set speed
; Parameter: $80-$8f: set global volume
; Parameter: $90-$9f: set filter mode.
;   bit 0: low pass, bit 2: band pass, bit 3: high pass,
;   bit 3: cut off voice 3's output.
; Parameter: $a0-$af: fine slide down.
; Parameter: $b0-$bf: fine slide up.
; Parameter: $c0-$cf: note cut.
; Parameter: $d0-$df: note delay.                               !!!! TODO
; Parameter: $e0-$ef: select filter controller
; Parameter: $f0-$ff: set hard restart ticks.
; ======================================
        lda speedcounter
        beq effect0finit
        jmp effect0frun
effect0finit:
        lda chn_effectpar,x
        bpl effect0fspeed
        and #$f0
        tay
        lda chn_effectpar,x
        and #$0f
; Now A is parameter bits 0..3, Y is bits 4..7
        cpy #$80
        beq effect0f80
        cpy #$90
        beq effect0f90
        cpy #$a0
        beq effect0fa0
        cpy #$b0
        beq effect0fb0
        cpy #$c0
        beq effect0fc0
        cpy #$d0
        beq effect0fd0
        cpy #$e0
        beq effect0fe0
        cpy #$f0
        beq effect0ff0
        rts
; Set speed.
effect0fspeed:
        sta speed
        rts
; Set global volume.
effect0f80:
        sta globalvolume
        rts
; Set filter mode.
effect0f90:
        asl
        asl
        asl
        asl
        sta filter_mode
        rts
; Fine slide down.
effect0fa0:
; Multiply parameter by 4 to get slide adder.
        asl
        asl
        sta player_vibratotemp
; Subtract from slidefreq.
        lda chn_slidefreqlo,x
        sec
        sbc player_vibratotemp
        sta chn_slidefreqlo,x
        lda chn_slidefreqhi,x
        sbc #$00
        jmp effect0fb0finish            ; short jump
; Fine slide up.
effect0fb0:
; Multiply parameter by 4 to get slide adder.
        asl
        asl
; Add to slidefreq.
        adc chn_slidefreqlo,x
        sta chn_slidefreqlo,x
        lda chn_slidefreqhi,x
        adc #$00
effect0fb0finish:
        sta chn_slidefreqhi,x
; Indicate that slide is active.
        lda #$80
        sta chn_slideactive,x
        rts
; Note cut.
effect0fc0:
        cmp speedcounter
        bne effect0fc0done
        lda #$fe                        ; Release gate.
        sta chn_gateon,x
effect0fc0done:
        rts
; Note delay. !!!! TODO
effect0fd0:
        rts
; Select filter cutoff controller.
effect0fe0:
        tay
        beq effect0fe0nofilter
        lda INSTRUMENTS_FILTERTABLESTART,y  ; Filter table start index.
        sta filter_idx
        lda INSTRUMENTS_FILTERTABLEEND,y ; Copy filter table end.
        sta filter_end
        lda INSTRUMENTS_FILTERTABLELOOP,y ; Copy filter table loop.
        sta filter_loop
        rts
effect0fe0nofilter:
        sta filter_idx
        sta filter_end
        sta filter_loop
        rts
; Set hardrestart.
effect0ff0:
        sta chn_hardrestart,x
        rts
effect0frun:
        lda chn_effectpar,x
        and #$f0
        tay
        lda chn_effectpar,x
        and #$0f
; Now A is parameter bits 0..3, Y is bits 4..7
        cpy #$c0
        beq effect0fc0
        rts

; ======================================
start_newpattern:
; Get order index -> pattern number -> tracks for channels 1,2,3.
; ======================================
;  Input: ordernumber
; Output: trackrow: = previous firsttrackrow
;         firsttrackrow: always 0
;         forcenewpattern: always 0
; ======================================
; about 93+3*71 clock cycles
        ldy ordernumber                 ; Get index in order table.
        lda player_singlepattern
        bne snp_orderready
        ldy nextordernumber             ; Get index in order table.
        sty ordernumber
        inc nextordernumber
snp_orderready:
; player_patternptr=PATTERNS+6*ORDERLIST[ordernumber]
        ldx #0
        lda ORDERLIST,y            ; Get current pattern.
        asl
        bcc snp_noof1
        inx
        clc
snp_noof1:
        adc ORDERLIST,y
        bcc snp_noof2
        inx
snp_noof2:
        stx player_patternptr+1
        asl
        rol player_patternptr+1
        adc #<PATTERNS
        sta player_patternptr
        lda player_patternptr+1
        adc #>PATTERNS
        sta player_patternptr+1
; Init row positions.
        lda firsttrackrow               ; Usually 0 or set by pattern break.
        sta trackrow
        lda #$00                        ; Default pattern start is row 0.
        sta firsttrackrow
        sta forcenewpattern             ; Do not force new pattern.
; Fetch transpose values and create track pointers for each channel.
        tax
        tay
        jsr start_newtrack              ; Get track for channel 1
        jsr start_newtrack              ; Get track for channel 2
        jmp start_newtrack              ; Get track for channel 3

; ======================================
start_newtrack:
; Read track number and calculate 192*track_number+tracks.
; Also read track transpose.
; ======================================
;  Input: player_patternptr, player_patternptr+1: pointer to pattern
;         X: channel number
;         Y: 2*channel number
; Output: chn_trackptrlo, chn_trackptrhi, chn_transpose
; ======================================
; about 71 clock cycles
        lda (player_patternptr),y              ; Get track number.
        asl
        adc (player_patternptr),y
        sta player_trackptr+1
        lda #$00
        ror player_trackptr+1
        ror
        lsr player_trackptr+1
        ror
        adc #<TRACKS_BASE
        sta chn_trackptrlo,x
        lda player_trackptr+1
        adc #>TRACKS_BASE
        sta chn_trackptrhi,x
        iny
        lda (player_patternptr),y              ; Get transpose.
        sta chn_transpose,x
        inx
        iny
        rts

; ======================================
look_newpattern:
; Get tracks for next pattern, but don's set track transpose.
; ======================================
;  Input: ordernumber
; ======================================
        ldy nextordernumber             ; Get index in order table.
; player_patternptr=PATTERNS+6*ORDERLIST[nextordernumber]
        ldx #0
        lda ORDERLIST,y            ; Get next pattern.
        asl
        bcc lnp_noof1
        inx
        clc
lnp_noof1:
        adc ORDERLIST,y
        bcc lnp_noof2
        inx
lnp_noof2:
        stx player_patternptr+1
        asl
        rol player_patternptr+1
        adc #<PATTERNS
        sta player_patternptr
        lda player_patternptr+1
        adc #>PATTERNS
        sta player_patternptr+1
; Create track pointers for each channel.
        ldx #0
        ldy #0
        jsr look_newtrack               ; Get track for channel 1
        jsr look_newtrack               ; Get track for channel 2
        jsr look_newtrack               ; Get track for channel 3
        rts

; ======================================
look_newtrack:
; Read track number and calculate 192*track_number+tracks.
; ======================================
;  Input: player_patternptr, player_patternptr+1: pointer to pattern
;         X: channel number
;         Y: 2*channel number
; Output: chn_trackptrlo, chn_trackptrhi
; ======================================
        lda (player_patternptr),y              ; Get track number.
        asl
        adc (player_patternptr),y
        sta player_trackptr+1
        lda #$00
        ror player_trackptr+1
        ror
        lsr player_trackptr+1
        ror
        adc #<TRACKS_BASE
        sta chn_trackptrlo,x
        lda player_trackptr+1
        adc #>TRACKS_BASE
        sta chn_trackptrhi,x
        inx
        iny
        iny
        rts

; ======================================
fetchrow:
; Parse a row of a track, read instrument data, etc.
; ======================================
;  Input: X: channel number
;         trackrow3
;         player_trackptr
; ======================================
; Fetch effect and parameter.
        ldy trackrow3
        iny
        lda (player_trackptr),y
        lsr
        lsr
        lsr
        lsr
        lsr
        sta chn_effect,x
        iny
        lda (player_trackptr),y
        sta chn_effectpar,x
; Fetch effect and instrument.
        dey
        dey
        lda (player_trackptr),y
        lsr
        lsr
        lsr
        lsr
        and #$08
        ora chn_effect,x
        sta chn_effect,x
        iny
        lda (player_trackptr),y
        and #$1f
        bne fr_newinst
        jmp fr_nonewinst
fr_newinst:
        sta chn_inst,x
        tay
; Copy changeable instrument parameters to channel.
        lda INSTRUMENTS_AD,y            ; Copy AD.
        sta chn_ad,x
        lda INSTRUMENTS_SR,y            ; Copy SR.
        sta chn_sr,x
        lda INSTRUMENTS_WAVETABLESTART,y ; Wave table start index.
        sta chn_waveidx,x
        lda INSTRUMENTS_ARPTABLESTART,y ; Arpeggio table start index.
        sta chn_arpidx,x
        lda INSTRUMENTS_VIBDELAY,y      ; Copy vibrato delay.
        sta chn_vibdelay,x
        lda INSTRUMENTS_PULSEWIDTH,y    ; Copy pulse width.
        jsr set_pulsewidth
        lda INSTRUMENTS_PULSESPEED,y    ; Copy pulse speed.
        sta chn_plsspeed,x
        lda INSTRUMENTS_PULSELIMITS,y   ; Copy pulse limits.
        jsr set_pulselimits
; Reset vibrato and pulse width modulation.
        lda #$00
        sta chn_plsdir,x
        sta chn_vibpos,x
fr_nonewinst:
; Fetch note.
        ldy trackrow3
        lda (player_trackptr),y
        and #$7f
        beq fr_nonewnote
        cmp #97
        beq fr_noteoff
; Normal note between C-0 and B-7.
        sec
        sbc #1
        sta chn_note,x
; If slide to note is active, set chn_notefreq?? only.
        ldy chn_effect,x
        cpy #3
        bne fr_noslidetonote
        clc
        adc chn_transpose,x
        jsr note2notefreq
; Restart wave and arpeggio table and vibrato if there is a new note.
fr_noslidetonote:
        ldy chn_inst,x
        lda INSTRUMENTS_WAVETABLESTART,y ; Wave table start index.
        sta chn_waveidx,x
        lda INSTRUMENTS_ARPTABLESTART,y ; Arpeggio table start index.
        sta chn_arpidx,x
; Reset slide.
        lda #$00
        sta chn_slidefreqlo,x
        sta chn_slidefreqhi,x
        sta chn_slideactive,x
; Set gate on.
        lda #$ff
        bne fr_setgate                  ; short jump
; Set gate off.
fr_noteoff:
        lda #$fe
fr_setgate:
        sta chn_gateon,x
fr_nonewnote:
        rts

; ======================================
set_pulsewidth:
; ======================================
;  Input: X: channel number
;         A: pulse width bits 4..11
; ======================================
        pha
        asl
        asl
        asl
        asl
        sta chn_plswidthlo,x
        pla
        lsr
        lsr
        lsr
        lsr
        sta chn_plswidthhi,x
        rts

; ======================================
set_pulselimits:
; ======================================
;  Input: X: channel number
;         A: pulse limits. bit 7..4 is lower limit, bit 3..0 is upper limit
;            of 4 MSBs of pulse width.
; ======================================
        pha
        lsr
        lsr
        lsr
        lsr
        sta chn_plslimitdown,x
        pla
        and #$0f
        sta chn_plslimitup,x
        rts

; ======================================
calcvibrato:
; Calc vibrato for channel X.
; ======================================
;  Input: X: channel number
;         chn_vibpos, chn_vibdepth
;         chn_freq??
; Output: chn_finfreq??
; ======================================
; Get vibrato frequency from table using depth.
        lda chn_vibpos,x
        and #$10
        beq cv_forward
        lda chn_vibpos,x
        eor #$0f
        jmp cv_gettable
cv_forward:
        lda chn_vibpos,x
cv_gettable:
        and #$0f
        ora chn_vibdepth,x              ; == depth*16
        tay
; vibrato_table,y is now the frequency that has to be
; added to or subtracted from the channel frequency.
        lda chn_vibpos,x
        and #$20
        bne cv_vibrato_addfreq
; chn_freq-=vibrato_table,y
        lda chn_freqlo,x
        sec
        sbc vibrato_table,y
        sta chn_finfreqlo,x
        lda chn_freqhi,x
        sbc #0
        sta chn_finfreqhi,x
        rts
cv_vibrato_addfreq:
; chn_notefreq+=vibrato_table,y
        lda chn_freqlo,x
        clc
        adc vibrato_table,y
        sta chn_finfreqlo,x
        lda chn_freqhi,x
        adc #0
        sta chn_finfreqhi,x
        rts

; ======================================
note2notefreq:
; Set freq for note of channel X from freq table.
; Set only chn_notefreq??, used when slide to note is done.
; ======================================
;  Input: X: channel number
;         A: note
; Output: chn_notefreqlo, chn_notefreqhi
; ======================================
        tay
        lda freqtablelo,y
        sta chn_notefreqlo,x
        lda freqtablehi,y
        sta chn_notefreqhi,x
        rts

; ======================================
note2freq:
; Set freq for note of channel X from freq table.
; ======================================
;  Input: X: channel number
;         A: note
; Output: chn_notefreqlo, chn_notefreqhi, chn_freqlo, chn_freqhi
; ======================================
        tay
        lda freqtablelo,y
        sta chn_notefreqlo,x
        sta chn_freqlo,x
        lda freqtablehi,y
        sta chn_notefreqhi,x
        sta chn_freqhi,x
        rts

; ======================================
freqcmp:
; See if chn_freq??<chn_notefreq?? for channel X. Used in slide to note.
; ======================================
;  Input: X: channel number
;         chn_freqlo, chn_freqhi: current frequency
;         chn_notefreqlo, chn_notefreqhi: destination frequency
; Output: FLAGS: C=!(chn_freq??<chn_notefreq??)
; ======================================
        lda chn_freqhi,x
        cmp chn_notefreqhi,x
        bne freqcmp_done
        lda chn_freqlo,x
        cmp chn_notefreqlo,x
freqcmp_done:
        rts

; Index of SID registers for each channel.
sidregindex:    .byte 0, 7, 14

; Pointers to effect procedures.
effectpointers: .word effect00, effect01, effect02, effect03
                .word effect04, effect05, effect06, effect07
                .word effect08, effect09, effect0a, effect0b
                .word effect0c, effect0d, effect0e, effect0f

; Frequencies for all notes.
        include freqtab/freqtab.s

; 1 table for each possible vibrato depth.
; Each table contains 1/4th of a 64 byte sine wave.
; Total 256 bytes.
vibrato_table:
        include vibrato/vibrato.s

; Channel data.
chn_muted:      .byte 0,0,0     ; Nonzero means channel is muted.
player_singlepattern: byte 0    ; If nonzero, play current pattern only.
playercleardata:                ; Variables from here are cleared when playing starts.
chn_hardrestart: .byte 0,0,0    ; Number of ticks to shut channel up before new note.
chn_gateon:     .byte 0,0,0     ; Bit 0 is SID gate bit, bit 7-1 always set.
chn_slideactive:.byte 0,0,0     ; Nonzero mean do slide.
chn_plsdir:     .byte 0,0,0     ; Nonzero means decrement pulse width.
chn_transpose:  .byte 0,0,0     ; Track transpose read from pattern.
chn_track:      .byte 0,0,0     ; Track index. Unused outside editor.
chn_trackptrlo: .byte 0,0,0     ; Track pointer low byte.
chn_trackptrhi: .byte 0,0,0     ; Track pointer high byte.
chn_note:       .byte 0,0,0     ; Note read from track.
chn_inst:       .byte 0,0,0     ; Instrument read from track.
chn_effect:     .byte 0,0,0     ; Effect read from track.
chn_effectpar:  .byte 0,0,0     ; Effect parameter read from track.
chn_waveform:   .byte 0,0,0     ; Waveform from wavetable or set waveform effect.
chn_ad:         .byte 0,0,0     ; CFI, modified by set AD effect. (CFI=Copy From Instrument)
chn_sr:         .byte 0,0,0     ; CFI, modified by set SR effect.
chn_plswidthlo: .byte 0,0,0     ; CFI, modified by set pulse width effect and pulse width modulation.
chn_plswidthhi: .byte 0,0,0     ; CFI, modified by set pulse width effect and pulse width modulation.
chn_plsspeed:   .byte 0,0,0     ; CFI, modified by set pulse speed effect.
chn_plslimitdown:.byte 0,0,0    ; CFI, modified by set pulse limits effect.
chn_plslimitup: .byte 0,0,0     ; CFI, modified by set pulse limits effect.
chn_vibdelay:   .byte 0,0,0     ; CFI, count down to 0 then start instrument vibrato.
chn_vibdepth:   .byte 0,0,0     ; CFI, modified by vibrato effect.
chn_vibspeed:   .byte 0,0,0     ; CFI, modified by vibrato effect.
chn_waveidx:    .byte 0,0,0     ; CFI, wave table index, updated every tick.
chn_arpidx:     .byte 0,0,0     ; CFI, arpeggio table index, updated every tick.
chn_notefreqlo: .byte 0,0,0     ; Frequency for note read from pattern.
chn_notefreqhi: .byte 0,0,0     ;
chn_freqlo:     .byte 0,0,0     ; Frequency after (arpeggio) and slide.
chn_freqhi:     .byte 0,0,0     ;   arpeggio is not added when slide to note.
chn_finfreqlo:  .byte 0,0,0     ; Final frequency after (arpeggio) and slide
chn_finfreqhi:  .byte 0,0,0     ;   and vibrato.
chn_vibpos:     .byte 0,0,0     ; Vibrato counter.
chn_slidefreqlo:.byte 0,0,0     ; New note resets this to 0,
chn_slidefreqhi:.byte 0,0,0     ;   slide effect adds/subs chn_slidefreq??.

; Global parameters.
filter_idx:     .byte 0         ; Filter table index.
filter_end:     .byte 0         ; Filter table end index.
filter_loop:    .byte 0         ; Filter table loop index.
filter_cutoff:  .byte 0         ; Filter cutoff frequency 8 MSBs.
filter_input:   .byte 0         ; Filter input and resonance.
filter_mode:    .byte 0         ; Filter mode as in $D418.
globalvolume:   .byte 0         ; Global volume that goes to SID.

PLAYERCLEARDATASIZE2 = *-playercleardata

; Variables describing the current position and status of the player.
mod3counter:    .byte 0         ; Mod 3 counter used by arpeggio effect
speed:          .byte 0         ; Speed as set by command F, default: 6
speedcounter:   .byte 0         ; Speed counter count from 0 to speed-1
trackrow:       .byte 0         ; Index of current trackrow.
trackrow3:      .byte 0         ; ==3*trackrow
nexttrackrow:   .byte 0         ; Next trackrow, used when looking ahead for hardrestart.
firsttrackrow:  .byte 0         ; First pattern row, usually 0 or set by pattern break.
ordernumber:    .byte 0         ; Index into orderlist.
nextordernumber:.byte 0         ; Next index into orderlist, set by order jump.
forcenewpattern:.byte 0         ; Flag set by order jump and pattern break.

PLAYERCLEARDATASIZE = *-playercleardata

editor_playerend:
