/*
  argv[1]: frequency of A-4 in Hz. (usually 440)
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// #define exactly 1 of these clock speeds.

#define FCLK    985248.0                // For PAL.
//#define FCLK    1022700.0               // For NTSC.
//#define FCLK    1000000.0               // For what? Atlantis maybe?

#ifndef M_LN2
#define M_LN2       0.693147180559945309417
#endif


long sidf[8*12];


void makeFreqTable(double a4)
{
  int o, n;

  double c7=exp(27.0*M_LN2/12.0)*a4;
  for (o=0; o<8; o++) {
    double od=1<<(7-o);
    for (n=0; n<12; n++) {
      double f=exp(n*M_LN2/12.0)*c7/od;
      sidf[o*12+n]=(long)(f*16777216.0/FCLK+0.5);
      if (sidf[o*12+n]>0xffff) sidf[o*12+n]=0xffff;     // Clip to 16 bit.
    }
  }

  printf("; Fclk=%.0lf Hz.\n\n", FCLK);

  printf("freqtablelo:\n");
  for (o=0; o<8; o++) {
    printf("        .byte");
    for (n=0; n<12; n++) {
      int t=(int)(sidf[o*12+n]&0xff);
      printf(" $%02x", t);
      if (n!=11) printf(",");
      else printf("        ; octave %d\n", o);
    }
  }

  printf("\nfreqtablehi:\n");
  for (o=0; o<8; o++) {
    printf("        .byte");
    for (n=0; n<12; n++) {
      int t=(int)(sidf[o*12+n]>>8);
      printf(" $%02x", t);
      if (n!=11) printf(",");
      else printf("        ; octave %d\n", o);
    }
  }
}


int main(int argc, char *argv[])
{
  double a4;

  sscanf(argv[1], "%lf", &a4);
  makeFreqTable(a4);
  return EXIT_SUCCESS;
}
